# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Tutorial on TreeBasedTensor.

'''

import numpy as np
import tensap

# %% Linear dimension tree
ORDER = 5
TREE = tensap.DimensionTree.linear(ORDER)
TREE.plot(title='Nodes indices')
TREE.plot_dims(title='Nodes dimensions')

# %% Random dimension tree
ORDER = 10
ARITY_INTERVAL = [2, 3]
TREE = tensap.DimensionTree.random(ORDER, ARITY_INTERVAL)
TREE.plot(title='Nodes indices')
TREE.plot_dims(title='Nodes dimensions')

# %% TreeBasedTensor: random generation
TENSOR = tensap.TreeBasedTensor.rand(TREE)
TENSOR.plot(title='Active nodes')
TENSOR.plot([x.storage() for x in TENSOR.tensors],
            title='Tensors\' storage complexity')
TENSOR.plot(TENSOR.representation_rank, title='Representation ranks')

# %% Truncation of a TreeBasedTensor with prescribed rank
TRUNCATOR = tensap.Truncator(tolerance=0, max_rank=np.random.randint(1, 5))
TENSOR = tensap.TreeBasedTensor.rand(TREE)
TENSOR_TRUNCATED = TRUNCATOR.hsvd(TENSOR)
print('Prescribed rank, error = %2.5e\n' %
      ((TENSOR - TENSOR_TRUNCATED).norm() / TENSOR.norm()))

# %% Truncation of a TreeBasedTensor with prescribed relative precision
ORDER = 10
TREE = tensap.DimensionTree.random(ORDER, 3)
TENSOR = tensap.TreeBasedTensor.rand(TREE)

TRUNCATOR = tensap.Truncator(tolerance=1e-8)
TRUNCATOR._hsvd_type = 1  # Root to leaves truncation
TENSOR_TRUNCATED_1 = TRUNCATOR.hsvd(TENSOR)
ERR_1 = (TENSOR - TENSOR_TRUNCATED_1).norm() / TENSOR.norm()
print('Root to leaves: prescribed tolerance = %2.5e, error = %2.5e\n' %
      (TRUNCATOR.tolerance, ERR_1))

TRUNCATOR._hsvd_type = 2  # Leaves to root truncation
TENSOR_TRUNCATED_2 = TRUNCATOR.hsvd(TENSOR)
ERR_2 = (TENSOR - TENSOR_TRUNCATED_2).norm() / TENSOR.norm()
print('Leaves to root: prescribed tolerance = %2.5e, error = %2.5e\n' %
      (TRUNCATOR.tolerance, ERR_2))

TENSOR_TRUNCATED_2.plot(TENSOR_TRUNCATED_2.representation_rank,
                        title='Representation ranks')

# %% Truncation of a FullTensor
ORDER = 5
ARITY_INTERVAL = [2, 3]
TREE = tensap.DimensionTree.random(ORDER, ARITY_INTERVAL)

SIZE = np.random.randint(8, 11, ORDER)
ACTIVE_NODES = tensap.TreeBasedTensor._random_is_active_node(TREE)

A = tensap.FullTensor.randn(SIZE)
TRUNCATOR = tensap.Truncator(tolerance=5e-1)
A_TRUNCATED = TRUNCATOR.hsvd(A, TREE, ACTIVE_NODES)
ERR = (A-A_TRUNCATED.full()).norm()/A.norm()
print('Tolerance = %2.5e, error = %2.5e\n' % (TRUNCATOR.tolerance, ERR))
A_TRUNCATED.plot(A_TRUNCATED.representation_rank,
                 title='Representation ranks')

TRUNCATOR.tolerance = 1e-2
A_TRUNCATED = TRUNCATOR.hsvd(A, TREE, ACTIVE_NODES)
ERR = (A-A_TRUNCATED.full()).norm()/A.norm()
print('Tolerance = %2.5e, error = %2.5e\n' % (TRUNCATOR.tolerance, ERR))
A_TRUNCATED.plot(A_TRUNCATED.representation_rank,
                 title='Representation ranks')

# %% Tensor train format
ORDER = 10
TREE = tensap.DimensionTree.linear(ORDER)
RANKS = np.hstack((1, np.random.randint(1, 3, TREE.nb_nodes-1)))

SIZE = np.full(ORDER, 3)
ACTIVE_NODES = np.logical_not(TREE.is_leaf)
ACTIVE_NODES[TREE.dim2ind[0]-1] = True
TENSOR = tensap.TreeBasedTensor.rand(TREE, RANKS, SIZE, ACTIVE_NODES)

TENSOR.plot(title='Nodes indices')
TENSOR.tree.plot_dims(title='Nodes dimensions')

# With a permutation of dimensions
RPERM = np.random.permutation(range(ORDER))
TREE = tensap.DimensionTree.linear(RPERM)
TENSOR = tensap.TreeBasedTensor.rand(TREE, RANKS, SIZE, ACTIVE_NODES)

TENSOR.plot(title='Nodes indices')
TENSOR.tree.plot_dims(title='Nodes dimensions')



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
#import os
#import sys
#import subprocess
# sys.path.insert(0, os.path.abspath('.'))


# -- Project information -----------------------------------------------------

project = 'tensap'
copyright = '2020, A. Nouy, E. Grelier'
author = 'A. Nouy, E. Grelier'

# The full version, including alpha/beta/rc tags
release = '1.1'


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = ['sphinx.ext.autodoc',
              'numpydoc',
              ]

#try:
    #import sphinx.ext.imgmath
    #extensions.append('sphinx.ext.imgmath')
    #imgmath_latex_preamble = r'\usepackage{{{0}math_notations}}'.format(
        #os.path.dirname(__file__) + os.sep)
    #imgmath_use_preview = True
    #if subprocess.call('dvisvgm -V', shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE) == 0:
        #imgmath_image_format = 'svg'
#except ImportError:
    #extensions.append('sphinx.ext.pngmath')
    #pngmath_latex_preamble = r'\usepackage{{{0}math_notations}}'.format(
        #os.path.dirname(__file__) + os.sep)
    ## The next option is used for smart-alignment of math images on the text.
    ## It only works when the preview-latex package is installed.
    ## See http://sphinx-doc.org/latest/ext/math.html#confval-pngmath_use_preview
    #pngmath_use_preview = True


autodoc_default_flags = ['members', 'inherited-members']
autosummary_generate = True
numpydoc_show_class_members = True
numpydoc_class_members_toctree = False

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'alabaster'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
#html_static_path = ['_static']

# -- Options for LATEX -------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    'papersize': 'a4paper',
    # The font size ('10pt', '11pt' or '12pt').
    'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    'preamble': r'\usepackage{math_notations}',
}

latex_additional_files = ['math_notations.sty']




# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Module uniform_random_variable.

'''

from numpy import array
from scipy.stats import uniform
from numpy.random import rand
import tensap


class UniformRandomVariable(tensap.RandomVariable):
    '''
    Class UniformRandomVariable.

    Attributes
    ----------
    moments : numpy.array
        The moments of the uniform random variable (if computed).
    inf : float, optional
        The lower bound of the support of the random variable. The default
        is -1.
    sup : float, optional
        The upper bound of the support of the random variable. The default
        is 1.

    '''

    def __init__(self, inf=-1, sup=1):
        '''
        Uniform random variable on [inf, sup].

        Parameters
        ----------
        inf : float, optional
            The lower bound of the support of the random variable. The default
            is -1.
        sup : float, optional
            The upper bound of the support of the random variable. The default
            is 1.

        Returns
        -------
        None.

        '''
        tensap.RandomVariable.__init__(self)

        self.inf = inf
        self.sup = sup

    def __repr__(self):
        return ('<{} on [{}, {}]>').format(self.__class__.__name__,
                                           self.inf,
                                           self.sup)

    def cdf(self, x):
        '''
        Evaluate the cumulative distribution function (cdf) of the uniform
        random variable at points x.

        Parameters
        ----------
        x : list or numpy.ndarray
            The points at which the cdf is to be evaluated.

        Returns
        -------
        numpy.ndarray
            The evaluations of the cdf.

        '''
        return uniform.cdf(x, self.inf, self.sup - self.inf)

    def icdf(self, x):
        '''
        Evaluate the inverse cumulative distribution function (icdf) of the
        uniform random variable at points x.

        Parameters
        ----------
        x : list or numpy.ndarray
            The points at which the icdf is to be evaluated.

        Returns
        -------
        numpy.ndarray
            The evaluations of the icdf.

        '''
        return uniform.ppf(x, self.inf, self.sup - self.inf)

    def pdf(self, x):
        '''
        Evaluate the probability density function (pdf) of the uniform
        random variable at points x.

        Parameters
        ----------
        x : list or numpy.ndarray
            The points at which the pdf is to be evaluated.

        Returns
        -------
        numpy.ndarray
            The evaluations of the pdf.

        '''
        return uniform.pdf(x, self.inf, self.sup - self.inf)

    def shift(self, bias, scaling):
        '''
        Shift the uniform random variable using the provided bias and scaling
        factor.

        The lower bound inf becomes scaling*inf + bias, and the upper bound
        sup becomes scaling*sup + bias.

        Parameters
        ----------
        bias : float
            The bias.
        scaling : float
            The scaling factor.

        Returns
        -------
        shifted_rv : tensap.UniformRandomVariable
            The shifted uniform random variable.

        '''
        shifted_rv = tensap.UniformRandomVariable(self.inf, self.sup)
        shifted_rv.inf = scaling * shifted_rv.inf + bias
        shifted_rv.sup = scaling * shifted_rv.sup + bias
        return shifted_rv

    @staticmethod
    def get_standard_random_variable():
        '''
        Return the standard uniform random variable on [-1, 1].

        Returns
        -------
        tensap.UniformRandomVariable
            The standard uniform random variable.

        '''
        return tensap.UniformRandomVariable()

    def support(self):
        '''
        Return the support of the uniform random variable.

        Returns
        -------
        numpy.ndarray
            Support of the uniform random variable.

        '''
        return array([self.inf, self.sup])

    def orthonormal_polynomials(self, *max_degree):
        '''
        Return the max_degree-1 first orthonormal polynomials associated with
        the UniformRandomVariable.

        Parameters
        ----------
        max_degree : int, optional
            The maximum degree of the returned polynomials. The default is
            None, choosing the default maximum degree associated with the
            constructor of the polynomials.

        Returns
        -------
        poly : tensap.OrthonormalPolynomials
            The generated orthonormal polynomials.

        '''
        poly = tensap.LegendrePolynomials(*max_degree)
        if self != UniformRandomVariable(-1, 1):
            # print('ShiftedOrthonormalPolynomials are created.')
            poly = tensap.ShiftedOrthonormalPolynomials(poly,
                                                        (self.inf+self.sup)/2,
                                                        (self.sup-self.inf)/2)
        return poly

    def get_parameters(self):
        '''
        Return the parameters of the uniform random variable.

        Returns
        -------
        float
            The lower bound of the support of the random variable.
        float
            The upper bound of the support of the random variable.

        '''
        return self.inf, self.sup

    def random_variable_statistics(self):
        return (self.inf + self.sup)/2, (self.sup - self.inf)**2/12

    def random(self, n=1):
        return rand(int(n)) * (self.sup - self.inf) + self.inf



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Tutorial on tree-based tensor approximation of multivariate probability density
functions using least-squares.

'''

from time import time
import numpy as np
from scipy.stats import multivariate_normal
import tensap

# %% Function to approximate: multivariate gaussian distribution
ORDER = 6
# Covariance matrix
S = np.array([[2,   0,   0.5, 1,   0,   0.5],
              [0,   1,   0,   0,   0.5, 0],
              [0.5, 0,   2,   0,   0,   1],
              [1,   0,   0,   3,   0,   0],
              [0,   0.5, 0,   0,   1,   0],
              [0.5, 0,   1,   0,   0,   2]])
# Reference measure
REF = tensap.RandomVector([tensap.UniformRandomVariable(-x, x)
                          for x in 5*np.diag(S)])


# Density, defined with respect to the reference measure
def U(x):
    return multivariate_normal.pdf(x, np.zeros(ORDER), S) * \
        np.prod(2*5*np.diag(S))


# %% Training and test samples
NUM_TRAIN = 1e5  # Training sample size
X_TRAIN = multivariate_normal.rvs(np.zeros(ORDER), S, int(NUM_TRAIN))

NUM_TEST = 1e4  # Test sample size
X_TEST = multivariate_normal.rvs(np.zeros(ORDER), S, int(NUM_TEST))

NUM_REF = 1e4  # Size of the sample used to compute the L2 error
XI = REF.random(NUM_REF)

# %% Approximation basis
DEGREE = 20
# Orthonormal bases in each dimension, with respect to the reference measure
BASES = [tensap.PolynomialFunctionalBasis(x.orthonormal_polynomials(),
                                          range(DEGREE+1)) for
         x in REF.random_variables]
BASES = tensap.FunctionalBases(BASES)

# %% Tree-based tensor learning parameters
SOLVER = tensap.TreeBasedTensorLearning.tensor_train_tucker(
        ORDER, tensap.DensityL2LossFunction())

RANDOMIZE = True
if RANDOMIZE:
    SOLVER.tree.dim2ind = np.random.permutation(SOLVER.tree.dim2ind)
    SOLVER.tree = SOLVER.tree.update_dims_from_leaves()

SOLVER.bases = BASES
SOLVER.bases_eval = BASES.eval(X_TRAIN)
SOLVER.training_data = X_TRAIN

SOLVER.tolerance['on_stagnation'] = 1e-6
# In density estimation, the error is the risk, which is negative
SOLVER.tolerance['on_error'] = -np.inf

SOLVER.initialization_type = 'random'

SOLVER.linear_model_learning.regularization = False
SOLVER.linear_model_learning.basis_adaptation = True
SOLVER.linear_model_learning.error_estimation = True

SOLVER.test_error = True
SOLVER.test_data = X_TEST
# SOLVER.bases_eval_test = BASES.eval(X_TEST)

SOLVER.rank_adaptation = True
SOLVER.rank_adaptation_options['max_iterations'] = 10
SOLVER.rank_adaptation_options['theta'] = 0.8
SOLVER.rank_adaptation_options['early_stopping'] = True
# early_stopping_factor < 1 because we the risk is negative
SOLVER.rank_adaptation_options['early_stopping_factor'] = 0.1

SOLVER.tree_adaptation = True
# For the tree adaptation in density estimation, a tolerance must be provided
SOLVER.tree_adaptation_options['tolerance'] = 1e-6

SOLVER.alternating_minimization_parameters['stagnation'] = 1e-6
SOLVER.alternating_minimization_parameters['max_iterations'] = 50

SOLVER.display = True
SOLVER.alternating_minimization_parameters['display'] = False

SOLVER.model_selection = True
SOLVER.model_selection_options['type'] = 'test_error'

# %% Density estimation
T0 = time()
F, OUTPUT = SOLVER.solve()
T1 = time()
print(T1-T0)

L2_ERR = np.linalg.norm(U(XI) - F(XI)) / np.linalg.norm(U(XI))
print('Risk leave-one-out estimation =       %2.5e' % OUTPUT['error'])
print('Risk estimation using a test sample = %2.5e' % OUTPUT['test_error'])
print('L2 relative error estimation =         %2.5e' % L2_ERR)

F.tensor.plot(title='Active nodes')
F.tensor.tree.plot_dims(title='Dimensions associated to the leaf nodes')
F.tensor.tree.plot_with_labels_at_nodes(F.tensor.representation_rank,
                                        title='Representation ranks')



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Module delta_functional_basis.

'''

from copy import deepcopy
import numpy as np
import tensap


class DeltaFunctionalBasis(tensap.FunctionalBasis):
    '''
    Class DeltaFunctionalBasis.

    Basis of univariate functions defined on a finite set of values in R^d.
    
    The is_orthonormal attribute remains at its default value of False.

    Attributes
    ----------
    values : numpy.ndarray of shape (n,d)
        contains the set of n values taken by the argument of the functions.
    measure : tensap.Measure
        The measure associated with the basis, by default a DiscreteMeasure. Can be a tensap.RandomVector or
        a tensap.RandomVariable to define a random generator and an
        expectation.

    '''

    def __init__(self, values=None, measure=None):
        '''
        Constructor for the class DeltaFunctionalBasis.

        The basis is not L2-orthonormal a priori, hence the is_orthonormal
        attribute remains at its default value of False.

        Parameters
        ----------
        values : numpy.ndarray of shape (n,d)
            contains the set of n values taken by the argument of the functions.
            The default is None.
        measure : tensap.Measure, optional
            The measure associated with the basis, by default a DiscreteMeasure. Can be a tensap.RandomVector or
            a tensap.RandomVariable to define a random generator and an
                expectation. The default is None.-

        Returns
        -------
        None.

        '''
        tensap.FunctionalBasis.__init__(self)
                   
        if values is not None:
            values = np.array(values)
            if values.ndim == 1:
                values = np.reshape(values, [-1, 1])
            self.values = values
        if measure is not None:
            self.measure = deepcopy(measure)
        elif values is not None:
            self.measure = tensap.DiscreteMeasure(values);
            
        if measure is not None:
            self.input_dimension = measure.ndim()
        else:
            self.input_dimension = 1

    def eval(self, x):
        x = np.atleast_1d(x)
        print(x)
        dim = self.values.shape[0]
        N = self.values.shape[0]
        out = np.zeros((np.shape(x)[0], dim))
        for mu in range(dim):
            for i in range(x.shape[0]):
                out[i, mu] = np.all(x[i,] == self.values[mu,])
        return out

    def domain(self):
        return self.measure.support()

    def ndim(self):
        return self.input_dimension

    def cardinal(self):
        return self.values.shape[1]



'''
Setup file for tensap (tensor approximation package).

Copyright (c) 2020, Anthony Nouy, Erwan Grelier
This file is part of tensap (tensor approximation package).

tensap is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

tensap is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="tensap",
    version="1.2",
    author="Anthony Nouy, Erwan Grelier",
    author_email="anthony.nouy@ec-nantes.fr",
    description="Tensor Approximation Package: a Python package for the approximation of functions and tensors.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://anthony-nouy.github.io/tensap/",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'numpy',
        'scipy',
    ],
    python_requires='>=3.6',
    include_package_data=True,
)



from tensap import chebyshev_points
from tensap import MultiIndices
from tensap import integer2baseb, baseb2integer

import numpy.testing as npt


def test_chebyshev_points():
    n = 10
    x = chebyshev_points(n, [-1, 1])
    print(x)
    npt.assert_almost_equal(x[0,0], 0.98768834)
    for i in range(n//2):
        npt.assert_almost_equal(x[i,0], -x[n-1-i, 0])


def test_multiindices():
    mi = MultiIndices.bounded_by([3, 5, 7])
    assert mi.cardinal() == 4*6*8, "wrong cardinal"
    mi_list = mi.to_list()
    assert len(mi_list) == mi.ndim(), "wrong list size"
    assert list(mi.get_indices(1)) == [1, 0, 0], "wrong indices"
    assert mi.is_downward_closed(), "wrong dw closed"


def test_integer2baseb():
    n = 85
    b = 3
    xb = integer2baseb(n, b)[0]
    xb_sum = sum([xb[-i-1] * b**i for i in range(len(xb))])
    assert n == xb_sum, "wrong int > base conversion"


def test_baseb2integer():
    I = [1,0,0,1,1]
    b = 3
    n = baseb2integer(I, b)
    assert n == 85, "wrong base > int conversion"



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Module canonical_tensor_learning.

'''

from copy import deepcopy
import numpy as np
import tensap


class CanonicalTensorLearning(tensap.TensorLearning):
    '''
    Class CanonicalTensorLearning.

    See also tensap.TensorLearning.

    Attributes
    ----------
    order : int
        The order of the tensor.

    '''

    def __init__(self, order, *args):
        '''
        Constructor for the class CanonicalTensorLearning.

        Parameters
        ----------
        order : int
            The order of the tensor.
        *args : tuple
            Additional parameters.

        Returns
        -------
        None.

        '''
        super().__init__(*args)
        self.order = order
        self.initialization_type = 'mean'
        self.truncate_initialization = True

# %% Standard solver methods
    def initialize(self):
        if self.tree_adaptation:
            print('tree_adaptation not defined for CanonicalTensorLearning.')
            self.tree_adaptation = False

        if 'one_by_one_factor' in self.alternating_minimization_parameters and\
            self.alternating_minimization_parameters['one_by_one_factor'] and \
                self.rank != 1:
            self._exploration_strategy = np.arange(
                1, self.alternating_minimization_parameters['inner_loops'] *
                self.rank*self.order+2)
            self._number_of_parameters = self._exploration_strategy.size
        else:
            self.alternating_minimization_parameters['one_by_one_factor'] = \
                False
            self._exploration_strategy = np.arange(1, self.order+2)
            self._number_of_parameters = self.order + 1

        shape = [x.shape[1] for x in self.bases_eval]
        if self.initialization_type == 'random':
            f = tensap.CanonicalTensor.randn(self.rank, shape)
        elif self.initialization_type == 'ones':
            f = tensap.CanonicalTensor.ones(self.rank, shape)
        elif self.initialization_type == 'initial_guess':
            f = self.initial_guess
        elif self.initialization_type == 'mean' or \
                self.initialization_type == 'mean_randomized':
            if not isinstance(self.training_data, list) or \
                    (isinstance(self.training_data, list) and
                     len(self.training_data) == 1):
                raise NotImplementedError('Initialization type not ' +
                                          'implemented in unsupervised ' +
                                          'learning.')
            if isinstance(self.bases, tensap.FunctionalBases):
                means = self.bases.mean()
            else:
                means = [np.mean(x, 0) for x in self.bases_eval]
            if self.initialization_type == 'mean_randomized':
                means = [x + 0.01*np.random.randn(*x.shape) for x in means]
            means = [np.reshape(x, [-1, 1]) for x in means]

            f = tensap.CanonicalTensor(
                means, np.atleast_1d(np.mean(self.training_data[1])))
        elif self.initialization_type == 'greedy':
            s_ini = deepcopy(self)
            s_ini.rank_adaptation = False
            s_ini.algorithm = 'greedy'
            s_ini.initialization_type = 'mean'
            s_ini.alternating_minimization_parameters['display'] = False
            s_ini.linear_model_learning.error_estimation = False
            s_ini.test_error = False
            s_ini.display = False
            f, output_ini = s_ini.solve()

            if self.display and 'error' in output_ini:
                print('Greedy initialization: rank = %i, error = %2.5e' %
                      (len(f.tensor.core.data), output_ini['error']))
        else:
            raise ValueError('Wrong initialization type.')

        if isinstance(f, tensap.FunctionalTensor):
            f = f.tensor

        if self.rank > len(f.core.data):
            fx = f.tensor_matrix_product(self.bases_eval).eval_diag().data

            s_ini = deepcopy(self)
            s_ini.rank_adaptation = False
            s_ini.algorithm = 'standard'
            s_ini.initialization_type = 'greedy'
            s_ini.rank = self.rank - len(f.core.data)
            s_ini.alternating_minimization_parameters['display'] = False
            s_ini.linear_model_learning.error_estimation = False
            s_ini.test_error = False
            s_ini.display = False
            if isinstance(s_ini.training_data, list) and \
                    len(s_ini.training_data) == 2:
                s_ini.training_data[1] -= fx
            elif isinstance(s_ini.loss_function, tensap.DensityL2LossFunction):
                s_ini.training_data = [s_ini.training_data, fx]

            f_add = s_ini.solve()[0]
            if isinstance(f_add, tensap.FunctionalTensor):
                f += f_add.tensor
            else:
                f += f_add

        if self.truncate_initialization:
            if self.order == 2:
                tr = tensap.Truncator()
                f = tr.truncate(f)
                f = tensap.CanonicalTensor(f.space, f.core.data)
                self.rank = len(f.core.data)
        return self, f

    def pre_processing(self, f):
        return self, f

    def randomize_exploration_strategy(self):
        if 'one_by_one_factor' not \
            in self.alternating_minimization_parameters or not \
                self.alternating_minimization_parameters['one_by_one_factor']:
            strategy = np.concatenate((
                np.random.permutation(self._number_of_parameters-1)+1,
                [self._number_of_parameters]))
        else:
            strat_mu = np.random.permutation(self.order)
            strat_i = np.random.permutation(self.rank)
            strategy = np.reshape(self._exploration_strategy[:-1],
                                  [-1, self.order], order='F')
            strategy[:, strat_mu] = np.array(strategy)
            strategy = np.reshape(np.reshape(strategy, [1, -1], order='F'),
                                  [self.rank, -1], order='F')
            strategy[strat_i, :] = np.array(strategy)
            strategy = np.concatenate((np.reshape(strategy, -1, order='F'),
                                       [self._number_of_parameters]))
        return strategy

    def prepare_alternating_minimization_system(self, f, mu):
        assert isinstance(self.loss_function, tensap.SquareLossFunction), \
            'Method not implemented for this loss function.'
        y = self.training_data[1]
        N = self.bases_eval[0].shape[0]

        if mu != self._number_of_parameters:
            if self.alternating_minimization_parameters['one_by_one_factor']:
                ind = mu
                mu = int(np.ceil(ind /
                                 self.alternating_minimization_parameters[
                                     'inner_loops'] / self.rank))
                ind = int(ind - self.rank * (np.ceil(ind / self.rank) - 1))

                coef = f.tensor.space[mu-1][:, ind-1]

                fH = f.tensor.tensor_matrix_product(self.bases_eval)

                fH_mu = np.ones((N, self.rank))
                no_mu = np.setdiff1d(np.arange(1, f.tensor.order+1), mu)
                for nu in no_mu:
                    fH_mu *= fH.space[nu-1]

                B = fH_mu * fH.space[mu-1]
                no_i = np.setdiff1d(np.arange(1, self.rank+1), ind)
                b = y - np.matmul(B[:, no_i-1], fH.core.data[no_i-1])

                A = self.bases_eval[mu-1] * np.transpose(
                    np.tile(fH_mu[:, ind-1], (f.tensor.shape[mu-1], 1)))
                if self.linear_model_learning[mu-1].basis_adaptation:
                    self.linear_model_learning[mu-1].basis_adaptation_path = \
                        self.bases_adaptation_path[mu-1]
            else:
                coef = f.tensor.space[mu-1]
                grad = f.parameter_gradient_eval(mu).transpose([0, 2, 1])
                A = np.reshape(grad.data, [grad.shape[0], -1], order='F')

                if self.linear_model_learning[mu-1].basis_adaptation:
                    self.linear_model_learning[mu-1].basis_adaptation_path = \
                        np.tile(self.bases_adaptation_path[mu-1],
                                (self.rank, 1))
                elif self.rank > 1:
                    self.linear_model_learning[mu-1].options[
                        'non_zero_blocks'] = np.empty(self.rank, dtype=object)
                    for kk in range(self.rank):
                        shape_mu = f.tensor.shape[mu-1]
                        self.linear_model_learning[mu-1].options[
                            'non_zero_blocks'][kk] = \
                            kk * shape_mu + np.arange(shape_mu)
                b = y
        else:
            coef = f.tensor.core.data
            if self.alternating_minimization_parameters['one_by_one_factor']:
                mu = self.order + 1
            grad = f.parameter_gradient_eval(mu)
            A = np.reshape(grad.data, [grad.shape[0], -1], order='F')
            b = y
            self.linear_model_learning[mu-1].basis_adaptation = False

        self.linear_model_learning[mu-1].initial_guess = np.reshape(coef, -1)

        return self, A, b, f

    def set_parameter(self, f, mu, coef):
        if mu != self._number_of_parameters:
            if not self.alternating_minimization_parameters[
                    'one_by_one_factor']:
                coef = np.reshape(coef, [f.tensor.shape[mu-1], self.rank],
                                  order='F')
                norm_coef = np.sqrt(np.sum(coef**2, 0))
                ind = norm_coef != 0
                if not np.all(ind):
                    print('Degenerate case: one factor is zero.')
                coef[:, ind] = np.matmul(coef[:, ind],
                                         np.diag(1/norm_coef[ind]))
                f.tensor.space[mu-1] = coef
                f.tensor.core.data = np.ravel(norm_coef)

                if len(f.tensor.space) == 2 and mu == 1:
                    f.tensor.space[0] = np.linalg.qr(
                        f.tensor.space[0])[0]
            else:
                ind = mu
                mu = int(np.ceil(ind /
                                 self.alternating_minimization_parameters[
                                     'inner_loops'] / self.rank))
                ind = int(ind - self.rank * (np.ceil(ind / self.rank) - 1))
                norm_coef = np.linalg.norm(coef)
                coef /= norm_coef
                f.tensor.space[mu-1][:, ind-1] = coef
                f.tensor.core.data[ind-1] = norm_coef
        else:
            f.tensor.core.data = coef
        return f

    def stagnation_criterion(self, f, f0):
        norm_f = f.tensor.norm()
        norm_f0 = f0.tensor.norm()
        return 2 * np.abs(norm_f - norm_f0) / (norm_f + norm_f0)

    def final_display(self, f):
        print('Rank = %i' % len(f.tensor.core.data), end='')

# %% Rank adaptation solver methods
    def local_solver(self):
        s_local = deepcopy(self)
        s_local.display = False
        s_local.rank_adaptation = False
        s_local.test_error = False
        s_local.algorithm = 'standard'
        s_local.initialization_type = 'mean'
        s_local.model_selection = False
        return s_local

    def new_rank_selection(self, f):
        return len(f.tensor.core.data) + 1, 1, deepcopy(f)

    def initial_guess_new_rank(self, s_local, f, *args):
        s_local.initialization_type = 'initial_guess'
        s_local.initial_guess = f
        return s_local

    def adaptation_display(self, f, *args):
        print('\tRank = %i' % len(f.tensor.core.data), flush=True)

# %% Greedy solver
    def _solve_greedy(self):
        '''
        Greedy solver in canonical tensor format.

        Raises
        ------
        ValueError
            If the number of LinearModelLearning objects is not equal to
            _numberOfParameters.

        Returns
        -------
        f : tensap.FunctionalTensor
            The learned approximation.
        output : dict
            The outputs of the solver.

        '''
        assert isinstance(self.loss_function, tensap.SquareLossFunction), \
            'Method not implemented for this loss function.'

        bases_eval = self.bases_eval
        output = {}
        output['sequence'] = np.empty(self.rank, dtype=object)

        # Replication of the LinearModelLearning objects
        if self.linear_model_learning_parameters[
                'identical_for_all_parameters'] and \
                not isinstance(self.linear_model_learning, (list, np.ndarray)):
            self.linear_model_learning = list(map(deepcopy, [
                self.linear_model_learning] * self._number_of_parameters))
        elif isinstance(self.linear_model_learning, (list, np.ndarray)) and \
                len(self.linear_model_learning) != self._number_of_parameters:
            raise ValueError('Must provide self._numberOfParameters ' +
                             'LinearModelLearning objects.')

        # Working set paths
        if isinstance(self.linear_model_learning, (list, np.ndarray)) and \
            np.any([x.basis_adaptation for
                    x in self.linear_model_learning]) and \
                self.bases_adaptation_path is None:
            self.bases_adaptation_path = self.bases.adaptation_path()

        y = self.training_data[1]

        s_local = deepcopy(self)
        s_local.algorithm = 'standard'
        s_local.rank = 1
        s_local.display = False
        s_local.test_error = False
        s_local.model_selection = False

        f = tensap.CanonicalTensor.zeros(0, [x.shape[1] for x in bases_eval])
        f_0 = deepcopy(f)
        stagnation = np.zeros((1, self.rank))

        ls_local = deepcopy(s_local.linear_model_learning
                            [self._number_of_parameters-1])

        is_error = False
        error = np.ones((1, self.rank))
        pure_greedy = False
        if not pure_greedy:
            for linear_solver in self.linear_model_learning:
                setattr(linear_solver, 'error_estimation', False)

        for k in np.arange(1, self.rank+1):
            s_local.training_data[1] = \
                y - f.tensor_matrix_product(bases_eval).eval_diag().data
            f_add, output_greedy = s_local.solve()
            if isinstance(f_add, tensap.FunctionalTensor):
                f += f_add.tensor
            else:
                f += f_add

            if not pure_greedy:
                f_H = f.tensor_matrix_product(bases_eval)
                A = np.ones((bases_eval[0].shape[0], len(f_H.core.data)))
                for space in f_H.space:
                    A *= space
                ls_local.basis_adaptation = False
                ls_local.basis = None
                ls_local.basis_eval = A
                ls_local.training_data = [None, y]
                coef, output_greedy = ls_local.solve()
                f.core.data = np.ravel(coef)

            stagnation[k-1] = 2*(f - f_0).norm() / (f.norm() + f_0.norm())
            current_rank = len(f.core.data)
            output['sequence'][k-1] = f

            if 'error' in output_greedy:
                error[k-1] = output_greedy['error']
                is_error = True
                if self.alternating_minimization_parameters['display']:
                    print('Alternating minimization (greedy): rank = %i, ' +
                          'error = %2.5e, stagnation = %2.5e', current_rank,
                          error[k-1], stagnation[k-1])
            else:
                if self.alternating_minimization_parameters['display']:
                    print('Alternating minimization (greedy): rank = %i, ' +
                          'stagnation = %2.5e', current_rank, stagnation[k-1])

            if error[k-1] < self.tolerance['on_error'] or \
                stagnation[k-1] < self.tolerance['on_stagnation'] or \
                    (k > 2 and error[k-1] > error[k-2] and
                     error[k-2] > error[k-3]):
                break

            f_0 = deepcopy(f)
            if self.test_error:
                f_eval_test = tensap.FunctionalTensor(f, self.bases_eval_test)
                output['test_error'] = self.loss_function.test_error(
                    f_eval_test, self.test_data)
                output['test_error_iterations'].append(output['test_error'])
                if self.display:
                    print('Greedy: iteration %i, test error = %2.5e', k,
                          output['test_error'])

        output['stagnation'] = stagnation[:k-1]
        if is_error:
            output['errors'] = error[:k-1]
            K = np.argmin(output['errors'])
            f = output['sequence'][K]
            output['selected_iterate'] = K
            output['error'] = output['errors'][K]

        if isinstance(self.bases, tensap.FunctionalBases):
            f = tensap.FunctionalTensor(f, self.bases)

        return f, output



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Tutorial on tensor truncation.

'''

import numpy as np
import tensap

# %% Definitions
ORDER = 6
NK = 8


def fun(inputs):
    '''
    Example function used for the tutorial.

    Parameters
    ----------
    inputs : numpy.ndarray
        The inputs points at which the function is to be evaluated.

    Returns
    -------
    numpy.ndarray
        The evaluations of the function at the input points.

    '''
    return 1/(3+inputs[:, 0]+inputs[:, 2]) + inputs[:, 1] + \
        np.cos(inputs[:, 3] + inputs[:, 4])


GRID = tensap.FullTensorGrid(np.linspace(0, 1, NK), ORDER)
ARRAY = GRID.array()

U = fun(ARRAY)
U = tensap.FullTensor(U, ORDER, np.full(ORDER, NK))

# %% Higher Order SVD - truncation in Tucker format
TR = tensap.Truncator()
TR.tolerance = 1e-8
UR = TR.hosvd(U)
print('Error = %2.5e' % ((UR.full()-U).norm()/U.norm()))
print('Storage = %d' % UR.storage())
print('Dimension of spaces = %s\n' % UR.tensors[0].shape)

# %% Tree-based format
ARITY_INTERVAL = [2, 3]
TREE = tensap.DimensionTree.random(ORDER, ARITY_INTERVAL)
TR = tensap.Truncator()
TR.tolerance = 1e-8
UR = TR.hsvd(U, TREE)

print('Error = %2.5e' % ((UR.full()-U).norm()/U.norm()))
print('Storage = %d' % UR.storage())
print('Ranks = %s\n' % UR.ranks)

# %% Truncation in tensor-train format
TR = tensap.Truncator()
TR.tolerance = 1e-8
UR = TR.ttsvd(U)
print(UR)

SIN_VAL = UR.singular_values()

print('Error = %2.5e' % ((UR.full()-U).norm()/U.norm()))
print('Storage = %d' % UR.storage())
print('TT-rank = %s\n' % UR.ranks)

# %% Truncation in tensor-train format after permutation
PERM = np.random.permutation(ORDER)
U_PERM = U.transpose(PERM)

TR = tensap.Truncator()
TR.tolerance = 1e-10
TR.max_rank = 200

UR = TR.ttsvd(U_PERM)
print('With permutation...')
print('\tOrdering = %s' % PERM)
print('\tError = %2.5e' % ((UR.full()-U_PERM).norm()/U_PERM.norm()))
print('\tStorage = %d' % UR.storage())
print('\tRanks = %s\n' % UR.ranks)

# %% Tensor-train tree optimization
print('Tree optimization...')
UR_PERM = UR.optimize_leaves_permutations(1e-10, 10)
print('\tInitial storage = %d' % UR.storage())
print('\tError = %2.5e' % ((UR_PERM.full()-U_PERM).norm() / U.norm()))
print('\tFinal storage = %d' % UR_PERM.storage())
print('\tRanks = %s' % UR_PERM.ranks)

UR.tree.plot_dims(title='Dimensions before tree optimization')
UR_PERM.tree.plot_dims(title='Dimensions before after optimization')



import glob
import pytest
import time
import matplotlib
import matplotlib.pyplot as plt
matplotlib.use('Agg')


def execfile(filepath, globals=None, locals=None):
    if globals is None:
        globals = {}
    globals.update({
        "__file__": filepath,
        "__name__": "__main__",
    })
    with open(filepath, 'rb') as file:
        exec(compile(file.read(), filepath, 'exec'), globals, locals)


@pytest.mark.parametrize("f", glob.glob('tutorials/**/*.py', recursive=True))
def test_tutorial(f):
    print(f"-- running {f}", flush=True)
    t0 = time.time()
    execfile(f)
    dt = time.time() - t0
    print(f"-- {f} ran in {dt:.2f} s", flush=True)
    plt.close("all")



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Module integration_rules.

'''

import numpy as np
import tensap


class IntegrationRule:
    '''
    Class IntegrationRule.

    Attributes
    ----------
    points : numpy.ndarray
        The integration points.
    weights : numpy.ndarray
        The integration weights.

    '''

    def __init__(self, points, weights):
        '''
        Constructor for the class IntegrationRule.

        Parameters
        ----------
        points : list or numpy.ndarray
        The integration points.
        weights : list or numpy.ndarray
            The integration weights.

        Returns
        -------
        None.

        '''
        self.points = np.atleast_1d(points)
        self.weights = np.atleast_1d(weights)

    def ndim(self):
        '''
        Return the number of integration points.

        Returns
        -------
        int
            The number of integration points.

        '''
        return self.points.shape[1]

    def integrate(self, fun):
        '''
        Integrate the function fun using the integration rule.

        Parameters
        ----------
        fun : function
            The function to be integrated.

        Returns
        -------
        int
            The integrated function using the integration rule.

        '''
        fun_eval = fun(self.points)
        return np.dot(np.ravel(fun_eval), self.weights)

    def tensorize(self, dim):
        '''
        Create a tensap.FullTensorProductIntegrationRule in dimension dim
        using the object.

        Parameters
        ----------
        dim : int
            The dimension of the tensap.FullTensorProductIntegrationRule.

        Returns
        -------
        tensap.FullTensorProductIntegrationRule
            The integration rule in dimension dim.

        '''
        points = tensap.FullTensorGrid(self.points, dim)
        weights = [self.weights]*dim
        return tensap.FullTensorProductIntegrationRule(points, weights)

    @staticmethod
    def gauss(random_variable, *args):
        '''
        Call the method gauss_integration_rule of tensap.RandomVariable or
        tensap.RandomVector.

        Parameters
        ----------
        random_variable : tensap.RandomVariable or tensap.RandomVector
            The random variable or vector associated to which the integration
            rule is to be computed.
        *args : misc
            Additional arguments.

        Returns
        -------
        tensap.IntegrationRule
            The integration rule associated with the random variable or vector.

        '''
        return random_variable.gauss_integration_rule(*args)


class FullTensorProductIntegrationRule(IntegrationRule):
    '''
    Class FullTensorProductIntegrationRule.

    '''

    def __init__(self, points, weights):
        tensap.IntegrationRule.__init__(self, points, weights)

        if isinstance(points, list):
            points = tensap.FullTensorGrid(points)
        elif not isinstance(points, tensap.FullTensorGrid):
            raise ValueError('The points must be a FullTensorGrid or a list.')

        assert isinstance(weights, list) and len(weights) == points.ndim(), \
            'The weights must be a list of length the length of points.'

        self.points = points
        self.weights = [np.ravel(x) for x in weights]

    def ndim(self):
        return self.points.ndim()

    def integrate(self, fun):
        points = self.points.array()
        weights = self.weights_on_grid()
        try:
            f_x = fun.eval(points)
        except Exception:
            try:
                f_x = fun(points)
            except Exception:
                raise ValueError('The function must be evaluable.')

        return np.dot(weights.T , f_x)

    def weights_on_grid(self):
        '''
        Return the weights on the grid.

        Returns
        -------
        numpy.ndarray
            The weights on the grid.

        '''
        weights = tensap.CanonicalTensor([np.reshape(x, [-1, 1]) for
                                          x in self.weights], [1])
        return np.ravel(weights.full().numpy())



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Tutorial on DimensionTree.

'''

import tensap

# %% Linear dimension tree
ORDER = 5
TREE = tensap.DimensionTree.linear(ORDER)
TREE.plot(title='Nodes indices')
TREE.plot_dims(title='Nodes dimensions')

# %% Random dimension tree
ORDER = 10
ARITY_INTERVAL = [2, 3]
TREE = tensap.DimensionTree.random(ORDER, ARITY_INTERVAL)
TREE.plot(title='Nodes indices')
TREE.plot_dims(title='Nodes dimensions')

# %% Balanced dimension tree
ORDER = 10
TREE = tensap.DimensionTree.balanced(ORDER)
TREE.plot(title='Nodes indices')
TREE.plot_dims(title='Nodes dimensions at the leaves')

TREE.plot(colored_nodes=TREE.ascendants(4), node_color='b',
          title='Ascendants of node 4')

TREE.plot(colored_nodes=TREE.descendants(4), node_color='b',
          title='Descendants of node 4')

# %% Extraction of a subtree
[SUB_TREE, NOD] = TREE.sub_dimension_tree(4)
TREE.plot(colored_nodes=NOD, node_color='b', title='Sub-tree to extract')
SUB_TREE.plot(title='Sub-tree')

# %% Plot with and without the same level at the same height
ORDER = 8
TREE = tensap.DimensionTree.random(ORDER)

TREE.plot_options['level_alignment'] = False
TREE.plot_with_labels_at_nodes(TREE.level, title='No level alignment')
TREE.plot_options['level_alignment'] = True
TREE.plot_with_labels_at_nodes(TREE.level, title='Level alignment')



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Tutorial on magic points and interpolation.

'''

import numpy as np
import matplotlib.pyplot as plt
import tensap

# %% Magic points associated with a Functional Basis (1D)
P1 = tensap.PolynomialFunctionalBasis(tensap.LegendrePolynomials(), range(21))
DOM = P1.domain()
G1 = np.linspace(DOM[0], DOM[1], 200)
MAGIC_POINTS, IND, OUTPUT = P1.magic_points(G1)

plt.figure()
plt.plot(G1, np.zeros(G1.shape), 'k.', markersize=1)
plt.plot(MAGIC_POINTS, np.zeros(MAGIC_POINTS.shape), 'ro', fillstyle='none')
plt.show()

# %% Magic points associated with a FullTensorProductFunctionalBasis
# Tensorization of 1D uniform grids for the selection of magic points in
# dimension d
D = 2
P1 = tensap.PolynomialFunctionalBasis(tensap.LegendrePolynomials(), range(21))
BASES = tensap.FunctionalBases.duplicate(P1, D)
P = tensap.FullTensorProductFunctionalBasis(BASES)
G1 = np.linspace(DOM[0], DOM[1], 30)
G = tensap.FullTensorGrid(G1, D).array()
MAGIC_POINTS, IND, OUTPUT = P.magic_points(G)

if D == 2:
    plt.figure()
    plt.plot(G[:, 0], G[:, 1], 'k.', markersize=1)
    plt.plot(MAGIC_POINTS[:, 0], MAGIC_POINTS[:, 1], 'ro', fillstyle='none')
    plt.show()

# %% Magic points associated with a SparseTensorProductFunctionalBasis
# Selection of magic points in dimension d in
# - a tensorization of 1D uniform grids
# - or a tensorization of 1D magic points
TENSORIZATION_OF_MAGIC_POINTS = True
D = 2
P1 = tensap.PolynomialFunctionalBasis(tensap.LegendrePolynomials(), range(21))
BASES = tensap.FunctionalBases.duplicate(P1, D)
W = [1, 2]  # Weights for the anisotropic sparsity
IND = tensap.MultiIndices.with_bounded_weighted_norm(2, 1, P1.cardinal()-1, W)
P = tensap.SparseTensorProductFunctionalBasis(BASES, IND)

if not TENSORIZATION_OF_MAGIC_POINTS:
    G1 = np.linspace(DOM[0], DOM[1], 100)
    G = tensap.FullTensorGrid(G1, D).array()
else:
    G1 = np.linspace(DOM[0], DOM[1], 1000)
    M1 = P1.magic_points(G1)[0]
    G = tensap.FullTensorGrid(np.ravel(M1), D).array()

MAGIC_POINTS, IND, OUTPUT = P.magic_points(G)

if D == 2:
    plt.figure()
    plt.plot(G[:, 0], G[:, 1], 'k.', markersize=1)
    plt.plot(MAGIC_POINTS[:, 0], MAGIC_POINTS[:, 1], 'ro', fillstyle='none')
    plt.show()

# %% Interpolation of a function using magic points
D = 2


def F(x):
    return np.cos(x[:, 0]) + x[:, 1]**6 + x[:, 0]**2*x[:, 1]


IF = P.interpolate(F, MAGIC_POINTS)

X = tensap.RandomVector(tensap.UniformRandomVariable(), D)
N_TEST = 1000
X_TEST = X.random(N_TEST)
ERR = np.linalg.norm(IF(X_TEST) - F(X_TEST)) / np.linalg.norm(F(X_TEST))
print('Test error = %2.5e' % ERR)



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Module orthonormal_polynomials.

'''

from abc import abstractmethod
import numpy as np
import tensap


class OrthonormalPolynomials(tensap.UnivariatePolynomials):
    '''
    Class OrthonormalPolynomials.

    Attributes
    ----------
    measure : tensap.Measure
        The measure associated with the orthonormal polynomials.

    '''

    @abstractmethod
    def __init__(self):
        '''
        Constructor for the class OrthonormalPolynomials.

        Returns
        -------
        None.

        '''
        self.measure = None
        self._recurrence_coefficients = np.array([])
        self._orthogonal_polynomials_norms = np.array([])

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    @staticmethod
    def is_orthonormal():
        return True

    @staticmethod
    def one():
        return 1, 0

    def __eq__(self, poly_2):
        if not isinstance(poly_2, OrthonormalPolynomials):
            out = False
        else:
            out = (self.measure == poly_2.measure and
                   np.array_equal(self._recurrence_coefficients,
                                  poly_2._recurrence_coefficients) and
                   np.array_equal(self._orthogonal_polynomials_norms,
                                  poly_2._orthogonal_polynomials_norms) and
                   self.is_orthonormal == poly_2.is_orthonormal)
        return out

    def domain(self):
        '''
        Return the support of the associated measure.

        Returns
        -------
        numpy.ndarray
            The support of the associated measure.

        '''
        return self.measure.support()

    def truncated_domain(self):
        '''
        Return the truncated support of the associated measure.

        Returns
        -------
        numpy.ndarray
            The truncated support of the associated measure.

        '''
        return self.measure.truncated_support()

    def moment(self, ind, measure=None):
        ind = np.atleast_2d(ind)

        if np.isin(ind.shape[1], [1, 2]) and (measure is None or
                                              measure == self.measure):
            out = np.zeros(ind.shape[0])
            if ind.shape[1] == 1:
                out[ind[:, 0] == 0] = 1
            else:
                out[ind[:, 0] == ind[:, 1]] = 1
        else:
            out = super().moment(ind, measure)
        return out

    def poly_coeff(self, ind):
        max_ind = np.max(ind)
        recurr = self._recurrence_coefficients

        if max_ind >= recurr.shape[1]:
            raise ValueError('Can generate polynomials up to degree{}.'
                             .format(recurr.shape[1]-1))

        coef = np.zeros([max_ind+1]*2)
        coef[0, 0] = 1

        if max_ind > 0:
            coef[1, 1:] = coef[0, :-1]
            coef[1, :] -= recurr[0, 0] * coef[0, :]

        if max_ind > 1:
            for deg in np.arange(1, max_ind):
                coef[deg+1, 1:] = coef[deg, :-1]
                coef[deg+1, :] -= recurr[0, deg] * coef[deg, :] + \
                    recurr[1, deg] * coef[deg-1, :]

        coef /= np.tile(np.expand_dims(
            self._orthogonal_polynomials_norms[:max_ind+1], axis=1),
                        (1, coef.shape[0]))
        return coef[ind, :]

    def polyval(self, ind, x):
        x = np.ravel(x)
        max_ind = np.max(ind)
        recurr = self._recurrence_coefficients

        if max_ind >= recurr.shape[1]:
            raise ValueError('Can generate polynomials up to degree {}.'
                             .format(recurr.shape[1]-1))

        recurr_1 = recurr[0, :max_ind+1]
        recurr_2 = recurr[1, :max_ind+1]
        norms = self._orthogonal_polynomials_norms[:max_ind+1]

        out = np.zeros((x.size, max_ind+1))
        out[:, 0] = 1
        if max_ind > 0:
            out[:, 1] = x - recurr_1[0]
            for deg in np.arange(2, max_ind+1):
                out[:, deg] = (x - recurr_1[deg-1]) * out[:, deg-1] - \
                    recurr_2[deg-1] * out[:, deg-2]
        out /= np.tile(norms[:max_ind+1], (x.size, 1))
        return out[:, ind]

    def d_polyval(self, ind, x):
        return self.dn_polyval(1, ind, x)

    def dn_polyval(self, n, ind, x):
        x = np.ravel(x)
        max_ind = np.max(ind)
        recurr = self._recurrence_coefficients
        recurr_1 = recurr[0, :max_ind+1]
        recurr_2 = recurr[1, :max_ind+1]
        norms = self._orthogonal_polynomials_norms[:max_ind+1]

        out = self.polyval(np.arange(max_ind+1), x) * np.tile(norms,
                                                              (x.size, 1))

        for n_tmp in np.arange(1, n+1):
            out_0 = np.array(out)

            out = np.zeros((x.size, max_ind+1))
            out[:, 0] = 0
            if n_tmp == 1:
                out[:, 1] = 1
            else:
                out[:, 1] = 0
            for deg in np.arange(2, max_ind+1):
                out[:, deg] = n_tmp * out_0[:, deg-1] + \
                    (x - recurr_1[deg-1]) * out[:, deg-1] - \
                    recurr_2[deg-1] * out[:, deg-2]
        out /= np.tile(norms[:max_ind+1], (x.size, 1))
        return out[:, ind]

    def random(self, ind, n=1, measure=None):
        '''
        Return an array of size n of random evaluations of the polynomials for
        which the degree is in ind. If measure is not provided, the
        random generation is performed using self.measure.

        Parameters
        ----------
        ind : list or numpy.ndarray
            The indices of the polynomials to be evaluated.
        n : int, optional
            The number of random evaluations. The default is 1.
        measure : tensap.ProbabilityMeasure, optional
            The probability measure used for the generation of the input
            points. The default is None, indicating to use
            self.measure.

        Returns
        -------
        out : numpy.ndarray
            The random evaluations of the polynomials.
        x : numpy.ndarray
            The randomly drawn input points.

        '''
        if measure is None:
            measure = self.measure

        assert isinstance(measure, tensap.ProbabilityMeasure), \
            'Must provide a ProbabilityMeasure.'
        n = np.atleast_1d(int(n))
        x = measure.random(np.prod(n))

        out = np.zeros((np.prod(n), len(ind)))
        for i, ind_loc in enumerate(ind):
            out[:, i] = self.polyval(ind_loc, x)
        if n.size > 1 and not (n.size == 2 and n[1] == 1):
            out = np.reshape(out, (-1, len(ind)))
            x = np.reshape(x, n)
        return out, x

    def roots(self, deg):
        '''
        Return the roots of the polynomial of degree deg.

        Parameters
        ----------
        deg : int
            The degree of the polynomial for which the roots are to be
            computed.

        Returns
        -------
        numpy.ndarray
            The roots of the polynomial of degree deg.

        '''
        try:
            coef = self._recurrence_coefficients[:, :deg]
        except AttributeError:
            coef = self.recurrence(self.measure, deg-1)

        # Jacobi matrix
        if deg == 1:
            jacobi_matrix = np.diag(coef[0, :])
        else:
            jacobi_matrix = np.diag(coef[0, :]) + \
                np.diag(np.sqrt(coef[1, 1:]), -1) + \
                np.diag(np.sqrt(coef[1, 1:]), 1)
        return np.sort(np.linalg.eig(jacobi_matrix)[0])


class ShiftedOrthonormalPolynomials(tensap.UnivariatePolynomials):
    '''
    Class ShiftedOrthonormalPolynomials.

    Attributes
    ----------
    measure : tensap.Measure
        The measure associated with the ShiftedOrthonormalPolynomials.
    polynomials : tensap.OrthonormalPolynomials
        The OrthonormalPolynomials which are shifted.
    shift : float
        The shifting parameter.
    scaling : float
        The scaling parameter.

    '''

    def __init__(self, polynomials, shift, scaling):
        '''
        Constructor for the class ShiftedOrthonormalPolynomials.

        Parameters
        ----------
        polynomials : tensap.OrthonormalPolynomials
            The OrthonormalPolynomials which are shifted.
        shift : float
            The shifting parameter.
        scaling : float
            The scaling parameter.

        Returns
        -------
        None.

        '''
        self.measure = polynomials.measure.shift(shift, scaling)
        self.polynomials = polynomials
        self.shift = shift
        self.scaling = scaling

    def __repr__(self):
        return '<{}>'.format(self.__class__.__name__)

    @staticmethod
    def is_orthonormal():
        return True

    @staticmethod
    def one():
        return 1, 0

    def __eq__(self, poly_2):
        if not isinstance(poly_2, ShiftedOrthonormalPolynomials):
            out = False
        else:
            out = (self.polynomials == poly_2.polynomials) and \
                (self.shift == poly_2.shift) and \
                (self.scaling == poly_2.scaling)
        return out

    def domain(self):
        '''
        Return the support of the associated shifted measure.

        Returns
        -------
        numpy.ndarray
            The support of the associated measure.

        '''
        return self.shift + self.scaling * self.polynomials.domain()

    def truncated_domain(self):
        '''
        Return the truncated support of the associated shifted measure.

        Returns
        -------
        numpy.ndarray
            The truncated support of the associated measure.

        '''
        return self.shift + self.scaling * self.polynomials.truncated_domain()

    def mean(self, ind, *measure):
        return self.polynomials.mean(ind, *measure)

    def moment(self, ind, *measure):
        if measure:
            if measure[0] == self.measure:
                measure = ()
            else:
                measure = tuple([measure[0].shift(-self.shift/self.scaling,
                                                  1/self.scaling)])
        return self.polynomials.moment(ind, *measure)

    def polyval(self, ind, x):
        x = (x - self.shift) / self.scaling
        return self.polynomials.polyval(ind, x)

    def d_polyval(self, ind, x):
        x = (x - self.shift) / self.scaling
        return self.polynomials.d_polyval(ind, x) / self.scaling

    def dn_polyval(self, n, ind, x):
        x = (x - self.shift) / self.scaling
        return self.polynomials.dn_polyval(ind, x) / (self.scaling**n)

    def random(self, ind, n=1, measure=None):
        '''
        Return an array of size n of random evaluations of the shifted
        polynomials for which the degree is in ind. If measure is not provided,
        the random generation is performed using self.measure.

        Parameters
        ----------
        ind : list or numpy.ndarray
            The indices of the polynomials to be evaluated.
        n : int, optional
            The number of random evaluations. The default is 1.
        measure : tensap.ProbabilityMeasure, optional
            The probability measure used for the generation of the input
            points. The default is None, indicating to use
            self.measure.

        Returns
        -------
        fx : numpy.ndarray
            The random evaluations of the polynomials.
        x : numpy.ndarray
            The randomly drawn input points.

        '''
        fx, x = self.polynomials.random(ind, n, measure)
        x = self.shift + self.scaling * x
        return fx, x

    def roots(self, n):
        '''
        Return the roots of the shifted polynomial of degree deg.

        Parameters
        ----------
        deg : int
            The degree of the polynomial for which the roots are to be
            computed.

        Returns
        -------
        numpy.ndarray
            The roots of the polynomial of degree deg.

        '''
        return self.shift + self.scaling * self.polynomials.roots(n)

    @staticmethod
    def ndim():
        return 1

    @staticmethod
    def poly_coeff():
        raise NotImplementedError('Method not implemented.')


class HermitePolynomials(OrthonormalPolynomials):
    '''
    Class HermitePolynomials.

    Polynomials defined on R and orthonormal with respect to the standard
    gaussian measure 1/sqrt(2*pi)*exp(-x^2/2).

    '''

    def __init__(self, n=50):
        '''
        Constructor for the class HermitePolynomials.

        Parameters
        ----------
        n : int, optional
            The highest degree for which a polynomial can be computed with the
            stored recurrence coefficients. The default is 50.

        Returns
        -------
        None.

        '''
        self.measure = tensap.NormalRandomVariable(0, 1)

        self._recurrence_coefficients, self._orthogonal_polynomials_norms = \
            self._recurrence(n)

    @staticmethod
    def _recurrence(n):
        '''
        Compute the coefficients of the three-term recurrence used to construct
        the Hermite polynomials, and the norms of the polynomials.

        The three-term recurrence writes:
        p_{n+1}(x) = (x-a_n)p_n(x) - b_n p_{n-1}(x), a_n and b_n are
        the three-term recurrence coefficients

        Parameters
        ----------
        n : int
            The highest degree for which a polynomial can be computed with the
            returned recurrence coefficients.

        Returns
        -------
        recurr : numpy.ndarray
            The recurrence coefficients.
        norms : numpy.ndarray
            The norms of the polynomials.

        '''
        recurr = np.vstack((np.zeros((1, n+1)), range(n+1)))
        norms = np.array([np.sqrt(np.float(np.math.factorial(x)))
                          for x in range(n+1)])
        return recurr, norms


class LegendrePolynomials(OrthonormalPolynomials):
    '''
    Class LegendrePolynomials.

    Polynomials defined on [-1,1], orthonormal with respect to the standard
    uniform measure 1/2.

    '''

    def __init__(self, n=50):
        '''
        Constructor for the class LegendrePolynomials.

        Parameters
        ----------
        n : int, optional
            The highest degree for which a polynomial can be computed with the
            stored recurrence coefficients. The default is 50.

        Returns
        -------
        None.

        '''
        self.measure = tensap.UniformRandomVariable(-1, 1)

        self._recurrence_coefficients, self._orthogonal_polynomials_norms = \
            self._recurrence(n)

    @staticmethod
    def _recurrence(n):
        '''
        Compute the coefficients of the three-term recurrence used to construct
        the Legendre polynomials, and the norms of the polynomials.

        The three-term recurrence writes:
        p_{n+1}(x) = (x-a_n)p_n(x) - b_n p_{n-1}(x), a_n and b_n are
        the three-term recurrence coefficients

        Parameters
        ----------
        n : int
            The highest degree for which a polynomial can be computed with the
            returned recurrence coefficients.

        Returns
        -------
        recurr : numpy.ndarray
            The recurrence coefficients.
        norms : numpy.ndarray
            The norms of the polynomials.

        '''
        recurr = np.vstack((np.zeros((1, n+1)),
                            np.arange(n+1)**2 / (4*np.arange(n+1)**2 - 1)))
        norms = np.array([np.sqrt(1/(2*x+1)) * 2**x * np.math.factorial(x)**2 /
                          np.math.factorial(2*x) for x in range(n+1)])
        return recurr, norms


class EmpiricalPolynomials(OrthonormalPolynomials):
    '''
    Class EmpiricalPolynomials.

    Polynomials defined on R and orthonormal with respect to the gaussian
    kernel smoothed distribution based on a sample x, which was centered and
    normalized (unit variance).

    '''

    def __init__(self, sample, n=None):
        '''
        Constructor for the class EmpiricalPolynomials.

        Parameters
        ----------
        sample : numpy.ndarray or list
            The sample used used to fit the probability density function using
            Scott's rule.
        n : int, optional
            The highest degree for which a polynomial can be computed with the
            stored recurrence coefficients. The default is 50.

        Returns
        -------
        None.

        '''
        if isinstance(sample, tensap.EmpiricalRandomVariable):
            self.measure = sample.get_standard_random_variable()
        else:
            # Standardization of the sample
            x = np.reshape(self.sample, [-1, 1])
            x = (x - np.tile(np.mean(x, 0), (x.shape[0], 1))) / \
                np.tile(np.std(x, 0), (x.shape[0], 1))
            self.measure = tensap.EmpiricalRandomVariable(np.ravel(x))

        self._recurrence_coefficients, self._orthogonal_polynomials_norms = \
            self._recurrence(self.measure, n)

    @staticmethod
    def _recurrence(measure, n):
        '''
        Compute the coefficients of the three-term recurrence used to construct
        the empirical polynomials, and the norms of the polynomials.

        The three-term recurrence writes:
        p_{n+1}(x) = (x-a_n)p_n(x) - b_n p_{n-1}(x), a_n and b_n are
        the three-term recurrence coefficients

        Parameters
        ----------
        measure : tensap.EmpiricalRandomVariable
            The empirical random variable associated with the orthonormal
            polynomials.
        n : int
            The highest degree for which a polynomial can be computed with the
            returned recurrence coefficients.

        Returns
        -------
        recurr : numpy.ndarray
            The recurrence coefficients.
        norms : numpy.ndarray
            The norms of the polynomials.

        '''

        def is_orth(pnp1, pn, pnm1, weights):
            '''
            Determine if the polynomial pnp1 is orthogonal to the polynomials
            pn and pnm1, according to the DiscreteRandomVariable r.

            Parameters
            ----------
            pnp1 : function
                The evaluations of the first function at integration points.
            pn : function
                The evaluations of the second function at integration points.
            pnm1 : function
                The evaluations of the third function at integration points.
            weights : numpy.ndarray
                The weights used for the numerical integration.

            Returns
            -------
            bool
                Boolean equal to True if the polynomial pnp1 is orthogonal to
                the polynomials pn and pnm1, False otherwise.

            '''
            tol = 1e-4  # Tolerance for the inner product to be considered as 0

            d1 = np.abs(np.sum(np.matmul(pnp1*pn, weights)))
            d2 = np.abs(np.sum(np.matmul(pnp1*pnm1, weights)))

            if d1 < tol and d2 < tol:
                return True
            return False

        def peval(i, a, b, x):
            '''
            Evaluate the polynomial of degree i defined by the coefficients a
            and b at points x.

            Parameters
            ----------
            i : int
                The degree of the polynomial.
            a : numpy.ndarray
                The first recurrence coefficients.
            b : numpy.ndarray
                The second recurrence coefficients.
            x : numpy.ndarray
                The points of evaluation.

            Returns
            -------
            p_n_loc : numpy.ndarray
                The evaluations of the polynomial at points x.

            '''

            if i < 0:
                p_n_loc = np.zeros(x.shape)
            elif i == 0:
                p_n_loc = np.ones(x.shape)
            else:
                p_n_loc_m2 = 1
                p_n_loc_m1 = x - a[0]
                p_n_loc = np.array(p_n_loc_m1)
                for N in np.arange(2, i+1):
                    p_n_loc = (x - a[N-1]) * p_n_loc_m1 - b[N-1] * p_n_loc_m2
                    p_n_loc_m2 = np.array(p_n_loc_m1)
                    p_n_loc_m1 = np.array(p_n_loc)
            return p_n_loc

        if n is None:
            check = True
            n = 10
        else:
            check = False

        norms = np.zeros(n+2)
        a = np.zeros(n+2)
        b = np.zeros(n+2)

        G = tensap.NormalRandomVariable().gauss_integration_rule(
            int(np.ceil((2*n+3)/2)))
        xi = measure.sample
        weights = G.weights / xi.size

        x_ij = measure.bandwidth * np.tile(np.reshape(G.points, [1, -1]),
                                           (xi.size, 1)) + \
            np.tile(np.reshape(xi, [-1, 1]), (1, G.points.size))

        i = 0
        cond = True
        norms[0] = 1
        a[0] = np.sum(np.matmul(x_ij, weights)) / xi.size
        b[0] = 0

        while cond and i <= n:
            i += 1

            p_n_m1 = np.reshape(peval(i-1, a, b, x_ij), x_ij.shape)
            p_n = np.reshape(peval(i, a, b, x_ij), x_ij.shape)

            norms[i] = np.sum(np.matmul(p_n**2, weights))
            a[i] = np.sum(np.matmul(p_n*x_ij*p_n, weights))/norms[i]
            b[i] = norms[i]/norms[i-1]

            p_n_p1 = (x_ij - a[i]) * p_n - b[i] * p_n_m1

            if check:
                # Orthogonality condition, only if the number of polynomials
                # is not specified by the user
                cond = is_orth(p_n_p1, p_n, p_n_m1, weights)

        if not cond and i-2 != n:
            raise ValueError('Maximum degree: %i (%i asked)' % (i-2, n))

        return np.vstack((a[:i], b[:i])), np.sqrt(norms[:i])


class DiscretePolynomials(OrthonormalPolynomials):
    '''
    Class DiscretePolynomials.

    Polynomials orthonormal with respect to a discrete measure.

    '''

    def __init__(self, measure=None):
        '''
        Constructor for the class DiscretePolynomials.

        Polynomials orthonormal with respect to a discrete measure.

        Parameters
        ----------
        measure : tensap.DiscreteRandomVariable
            The discrete measure with respect to which the polynomials form an
            orthonormal basis.

        Returns
        -------
        None.

        '''
        assert isinstance(measure, tensap.DiscreteRandomVariable), \
            'Must specify a DiscreteRandomVariable.'

        self.measure = measure
        n = np.size(measure.values)-1
        self._recurrence_coefficients, self._orthogonal_polynomials_norms = \
            self._recurrence(measure, n)

    @staticmethod
    def _recurrence(measure, n):
        '''
        Compute the coefficients of the three-term recurrence used to construct
        the discrete polynomials, and the norms of the polynomials.

        The three-term recurrence writes:
        p_{n+1}(x) = (x-a_n)p_n(x) - b_n p_{n-1}(x), a_n and b_n are
        the three-term recurrence coefficients

        Parameters
        ----------
        measure : tensap.DiscreteRandomVariable
            The discrete measure with respect to which the polynomials form an
            orthonormal basis.
        n : int
            The highest degree for which a polynomial can be computed with the
            returned recurrence coefficients.

        Returns
        -------
        recurr : numpy.ndarray
            The recurrence coefficients.
        norms : numpy.ndarray
            The norms of the polynomials.

        '''
        def dot_product(p1, p2, r):
            '''
            Compute the inner product between two polynomials p1 and p2,
            according to the DiscreteRandomVariable r.

            Parameters
            ----------
            p1 : function
                The first function of the inner product.
            p2 : function
                The second function of the inner product.
            r : tensap.DiscreteRandomVariable
                The random variable used to compute the inner product.

            Returns
            -------
            float
                The inner product between p1 and p2.

            '''
            G = r.integration_rule()
            return G.integrate(lambda x: p1(x)*p2(x))

        def is_orth(pnp1, pn, pnm1, r):
            '''
            Determine if the polynomial pnp1 is orthogonal to the polynomials
            pn and pnm1, according to the DiscreteRandomVariable r.

            Parameters
            ----------
            pnp1 : function
                The first function.
            pn : function
                The second function.
            pnm1 : function
                The third function.
            r : tensap.DiscreteRandomVariable
                The random variable used to compute the inner product.

            Returns
            -------
            bool
                Boolean equal to True if the polynomial pnp1 is orthogonal to
                the polynomials pn and pnm1, False otherwise.

            '''
            tol = 1e-5  # Tolerance for the inner product to be considered as 0

            d1 = np.abs(dot_product(pnp1, pn, r))
            d2 = np.abs(dot_product(pnp1, pnm1, r))

            if d1 < tol and d2 < tol:
                return True
            return False

        norms = np.zeros(n+2)
        a = np.zeros(n+2)
        b = np.zeros(n+2)

        i = 0
        cond = True
        norms[0] = 1
        a[0] = dot_product(lambda x: x**0, lambda x: x, measure)
        b[0] = 0
        pnm1 = []
        pn = [lambda x: 1]
        pnp1 = [lambda x, a=a[0]: (x-a)*pn[0](x)]

        while cond and i <= n:
            i += 1
            pnm1.append(pn[i-1])
            pn.append(pnp1[i-1])

            norms[i] = dot_product(pn[i], pn[i], measure)
            if norms[i] == 0:
                a[i] = np.nan
            else:
                a[i] = dot_product(pn[i],
                                   lambda x: x*pn[i](x), measure) / norms[i]
            b[i] = norms[i] / norms[i-1]

            pnp1.append(lambda x, a=a[i], b=b[i], pn=pn[i], pnm1=pnm1[i-1]:
                        (x-a)*pn(x) - b*pnm1(x))

            # Orthogonality condition
            cond = is_orth(pnp1[i], pn[i], pnm1[i-1], measure)

        return np.vstack((a[:i], b[:i])), np.sqrt(norms[:i])



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Module user_defined_functional_basis.

'''

from copy import deepcopy
import numpy as np
import tensap


class UserDefinedFunctionalBasis(tensap.FunctionalBasis):
    '''
    Class UserDefinedFunctionalBasis.

    The basis is not L2-orthonormal a priori, hence the is_orthonormal
    attribute remains at its default value of False.

    Attributes
    ----------
    handle_fun : numpy.ndarray
        The functions making the basis.
    measure : tensap.Measure
        The measure associated with the basis. Can be a tensap.RandomVector or
        a tensap.RandomVariable to define a random generator and an
        expectation.
    input_dimension : int
        The dimension of the domain of the functions in handle_fun.

    '''

    def __init__(self, h_fun=None, measure=None, input_dim=None):
        '''
        Constructor for the class UserDefinedFunctionalBasis.

        The basis is not L2-orthonormal a priori, hence the is_orthonormal
        attribute remains at its default value of False.

        Parameters
        ----------
        h_fun : list or numpy.ndarray, optional
            The functions making the basis. The default is None.
        measure : tensap.Measure, optional
            The measure associated with the basis. Can be a tensap.RandomVector
            or a tensap.RandomVariable to define a random generator and an
            expectation. The default is None.
        input_dim : input_dimension, optional
            The dimension of the domain of the functions in handle_fun. The
            default is None, indicating to deduce if from measure.ndim().

        Returns
        -------
        None.

        '''
        tensap.FunctionalBasis.__init__(self)
        self.handle_fun = np.ravel(h_fun)

        if measure is not None:
            self.measure = deepcopy(measure)
        if input_dim is None and measure is not None:
            self.input_dimension = measure.ndim()
        elif input_dim is not None:
            self.input_dimension = input_dim
        else:
            self.input_dimension = 1

    def eval(self, x):
        x = np.atleast_1d(x)
        dim = len(self.handle_fun)
        out = np.zeros((np.shape(x)[0], dim))
        for mu in range(dim):
            out[:, mu] = np.ravel(self.handle_fun[mu](x))
        return out

    def domain(self):
        return self.measure.support()

    def ndim(self):
        return self.input_dimension

    def cardinal(self):
        return self.handle_fun.size



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Module measure.

'''

from abc import ABC, abstractmethod


class Measure(ABC):
    '''
    Class Measure.

    '''

    @abstractmethod
    def __eq__(self, measure_2):
        pass

    @abstractmethod
    def mass(self):
        '''
        Return the mass of the Measure.

        Returns
        -------
        None.

        '''
        pass

    @abstractmethod
    def support(self):
        '''
        Return the support of the Measure.

        Returns
        -------
        None.

        '''
        pass

    @abstractmethod
    def ndim(self):
        '''
        Return the dimension of the Measure.

        Returns
        -------
        None.

        '''
        pass



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Initialization file for tensap (tensor approximation package).

'''

from .linear_algebra.magic_indices import magic_indices

from .tensor_algebra.tensors.dimension_tree import DimensionTree
from .tensor_algebra.tensors.full_tensor import FullTensor
from .tensor_algebra.tensors.tree_based_tensor import TreeBasedTensor
from .tensor_algebra.tensors.canonical_tensor import CanonicalTensor
from .tensor_algebra.tensors.diagonal_tensor import DiagonalTensor
from .tensor_algebra.tensors.sparse_tensor import SparseTensor
from .tensor_algebra.tools.truncator import Truncator

from .tools.utils import *
from .tools.multi_indices import MultiIndices
from .tools.grids import TensorGrid, FullTensorGrid, SparseTensorGrid
from .tools.chebyshev_points import chebyshev_points

from .approximation.integration.integration_rules import *

from .functions.measures.copulas import *
from .functions.measures.measure import Measure
from .functions.measures.lebesgue_measure import LebesgueMeasure
from .functions.measures.product_measure import ProductMeasure
from .functions.measures.probability_measure import ProbabilityMeasure
from .functions.measures.random_variable import RandomVariable
from .functions.measures.normal_random_variable import NormalRandomVariable
from .functions.measures.uniform_random_variable import UniformRandomVariable
from .functions.measures.empirical_random_variable import \
    EmpiricalRandomVariable
from .functions.measures.discrete_random_variable import DiscreteRandomVariable
from .functions.measures.discrete_measure import DiscreteMeasure
from .functions.measures.random_vector import RandomVector
from .functions.measures.random_multi_indices import random_multi_indices
from .functions.polynomials.polynomials import *
from .functions.polynomials.orthonormal_polynomials import *
from .functions.sets.is_in import is_in
from .functions.function import Function
from .functions.functional_tensor import FunctionalTensor
from .functions.user_defined_function import UserDefinedFunction
from .functions.tensorizer import Tensorizer
from .functions.tensorized_function import TensorizedFunction
from .functions.compositional_model_function import \
    CompositionalModelFunction
from .functions.multivariate_functions_benchmark import \
    multivariate_functions_benchmark

from .approximation.tools.model_selection import ModelSelection
from .approximation.bases.functional_basis_array import FunctionalBasisArray
from .approximation.bases.functional_bases import FunctionalBases
from .approximation.bases.functional_basis import FunctionalBasis
from .approximation.bases.polynomial_functional_basis import \
    PolynomialFunctionalBasis
from .approximation.bases.sub_functional_basis import SubFunctionalBasis
from .approximation.bases.user_defined_functional_basis import \
    UserDefinedFunctionalBasis
from .approximation.bases.full_tensor_product_functional_basis import \
        FullTensorProductFunctionalBasis
from .approximation.bases.sparse_tensor_product_functional_basis import \
        SparseTensorProductFunctionalBasis
from .approximation.bases.delta_functional_basis import DeltaFunctionalBasis

from .approximation.learning.loss_functions import *
from .approximation.learning.learning import Learning
from .approximation.learning.linear_model_learning import LinearModelLearning
from .approximation.learning.linear_model_learning_custom_loss import \
    LinearModelLearningCustomLoss
from .approximation.learning.linear_model_learning_square_loss import \
    LinearModelLearningSquareLoss
from .approximation.learning.linear_model_learning_density_l2 import \
    LinearModelLearningDensityL2
from .approximation.tensor_approximation.tensor_learning.tensor_learning \
    import TensorLearning
from .approximation.tensor_approximation.tensor_learning.\
    tree_based_tensor_learning import TreeBasedTensorLearning
from .approximation.tensor_approximation.tensor_learning.\
    canonical_tensor_learning import CanonicalTensorLearning
from .approximation.tensor_approximation.principal_component_analysis.\
    tensor_principal_component_analysis import TensorPrincipalComponentAnalysis
from .approximation.tensor_approximation.principal_component_analysis.\
    functional_tensor_principal_component_analysis import \
        FunctionalTensorPrincipalComponentAnalysis

# Disable tensorflow's warning messages
import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '1'

__pdoc__ = {"tutorials": False}



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Tutorial on Tensorizer and TensorizedFunction.

'''

import numpy as np
import tensap

# %% Identification of a function f(x) on (0,1) with a function g(i1, ...,id , y)
# x is identified with (i_1, ..., i_d, y) through a Tensorizer
# First interpolate and then truncate

L = 10  # Resolution
B = 3  # Scaling factor

T = tensap.Tensorizer(B, L, 1)

def FUN(x):
    return np.sqrt(x)
FUN = tensap.UserDefinedFunction(FUN, 1)
FUN.evaluation_at_multiple_points = True

TENSORIZED_FUN = T.tensorize(FUN)
TENSORIZED_FUN.fun.evaluation_at_multiple_points = True

# Interpolation of the function in the tensor product feature space 
DEGREE = 3
BASES = T.tensorized_function_functional_bases(DEGREE)
H = tensap.FullTensorProductFunctionalBasis(BASES)
FUN_INTERP, _ = H.tensor_product_interpolation(TENSORIZED_FUN)
TENSORIZED_FUN_INTERP = tensap.TensorizedFunction(FUN_INTERP, T)
X_TEST = T.X.random(100)
F_X_TEST = TENSORIZED_FUN_INTERP(X_TEST)
Y_TEST = FUN(X_TEST)
ERR_L2 = np.linalg.norm(Y_TEST - F_X_TEST) / np.linalg.norm(Y_TEST)
print('Mean squared error for the interpolation = %2.5e' % ERR_L2)

# Truncation in tensor train format
TR = tensap.Truncator()
tens = TENSORIZED_FUN_INTERP.fun.tensor
for k in range(1,9):
    TR.tolerance = 10**(-k)
    print('Tolerance =%s' % TR.tolerance)
    TENSORIZED_FUN_TT = TENSORIZED_FUN_INTERP
    TENSORIZED_FUN_TT.fun.tensor = TR.ttsvd(tens)
    print('Representation ranks = %s' % TENSORIZED_FUN_TT.fun.tensor.representation_rank)
    print('Complexity = %s' % TENSORIZED_FUN_TT.fun.tensor.storage())
    X_TEST = T.X.random(1000)
    F_X_TEST = TENSORIZED_FUN_TT(X_TEST)
    Y_TEST = FUN(X_TEST)
    ERR_L2 = np.linalg.norm(Y_TEST - F_X_TEST) / np.linalg.norm(Y_TEST)
    print('Mean squared error = %2.5e' % ERR_L2)

# %% Identification of a bivariate function f(x1, x2) on (0,1)^2 with a function
# g(i1,j1, ..., id,jd, y1, y2)
# x1 and x2 are identified with (i_1, ,..., i_d, y1) and (j_1, ...., j_d, y2)
# through a Tensorizer
# First interpolate and then truncate

DIM = 2
L = 8  # Resolution
B = 2  # Scaling factor

T = tensap.Tensorizer(B, L, DIM)
T.ordering_type = 2
FUN = tensap.UserDefinedFunction('1/(1+x0+x1)', DIM)
FUN.evaluation_at_multiple_points = True
TENSORIZED_FUN = T.tensorize(FUN)

# Interpolation of the function in the tensor product feature space
DEGREE = 2
BASES = T.tensorized_function_functional_bases(DEGREE)

H = tensap.FullTensorProductFunctionalBasis(BASES)
FUN_INTERP, _ = H.tensor_product_interpolation(TENSORIZED_FUN)
TENSORIZED_FUN_INTERP = tensap.TensorizedFunction(FUN_INTERP, T)
X_TEST = T.X.random(100)
F_X_TEST = TENSORIZED_FUN_INTERP(X_TEST)
Y_TEST = FUN(X_TEST)
ERR_L2 = np.linalg.norm(Y_TEST - F_X_TEST) / np.linalg.norm(Y_TEST)
print('Mean squared error for the interpolation = %2.5e' % ERR_L2)

# Truncation in tensor train format
TR = tensap.Truncator()
tens = TENSORIZED_FUN_INTERP.fun.tensor
for k in range(1,9):
    TR.tolerance = 10**(-k)
    print('Tolerance =%s' % TR.tolerance)
    TENSORIZED_FUN_TT = TENSORIZED_FUN_INTERP
    TENSORIZED_FUN_TT.fun.tensor = TR.ttsvd(tens)
    print('Representation ranks = %s' % TENSORIZED_FUN_TT.fun.tensor.representation_rank)
    print('Complexity = %s' % TENSORIZED_FUN_TT.fun.tensor.storage())
    X_TEST = T.X.random(1000)
    F_X_TEST = TENSORIZED_FUN_TT(X_TEST)
    Y_TEST = FUN(X_TEST)
    ERR_L2 = np.linalg.norm(Y_TEST - F_X_TEST) / np.linalg.norm(Y_TEST)
    print('Mean squared error = %2.5e' % ERR_L2)
 
    


# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Module tensorizer.

'''

import numpy as np
import tensap


class Tensorizer:
    '''
    Class Tensorizer.

    Attributes
    ----------
    b : int
        The base of the map.
    d : int
        The resolution of the map.
    dim : int
        The input dimension of the function to be tensorized.
    X : tensap.RandomVector
        Random vector used for the map.
    Y : tensap.RandomVector
        Random vector used for the map.

    '''

    def __init__(self, b, d, dim=1, X=None, Y=None):
        '''
        Constructor for the class Tensorizer.

        If dim == 1, defines a map t from [0,1] to {0,...,b-1}^d x [0,1]
        t(x) = (i_1, ..., i_d, y) with y in [0,1] and i_k in {0, ..., b-1}
        such that x = (i + y)b^(-d) with i in {0, ..., b^d-1}
        having the following representation in base b:
        i = sum_{k=1}^d i_k b^(d-k) in [0,b^d-1].

        If dim != 1, defines a map t from [0,1]^dim to
        {{0, ..., b-1}^d}^dim x [0,1]^dim if property orderingType == 1,
        or from [0,1]^dim to {{0, ..., b-1}^dim}^d x [0,1]^dim if property
        orderingType == 2.

        Parameters
        ----------
        b : int
        The base of the map.
        d : int
            The resolution of the map.
        dim : int, optional
            The input dimension of the function to be tensorized. The default
            is 1.
        X : tensap.RandomVector, optional
            Random vector used for the map. The default is None a uniform
            random vector on [0, 1]^dim.
        Y : tensap.RandomVector, optional
            Random vector used for the map. The default is None a uniform
            random vector on [0, 1]^dim.
        ordering_type : int
            Integer specifying the ordering type of the variables. 
            The default is 1. 

        Returns
        -------
        None.

        '''
        self.dim = dim
        self.b = b
        self.d = d

        self.ordering_type = 1

        if X is not None:
            if isinstance(X, tensap.RandomVariable):
                X = tensap.RandomVector(X, self.dim)
            self.X = X
        else:
            self.X = tensap.RandomVector(tensap.UniformRandomVariable(0, 1),
                                         self.dim)

        if Y is not None:
            if isinstance(Y, tensap.RandomVariable):
                Y = tensap.RandomVector(Y, self.dim)
            self.Y = Y
        else:
            self.Y = tensap.RandomVector(tensap.UniformRandomVariable(0, 1),
                                         self.dim)

    def map(self, x, nargout=1):
        '''
        Evaluate the map at points x, returning y and i such that
        self.map(x) = (i_1, ..., i_d, y).

        Parameters
        ----------
        x : list or numpy.ndarray
            The points at which the map is to be evaluated.
        nargout : int, optional
            The number of outputs. The default is 1, returning a horizontal
            stack of i and y. Set to 2 to return y and i separately.

        Returns
        -------
        numpy.ndarray
            Either (i_1, ..., i_d, y) if nargout == 1, or y if nargout == 2.
        numpy.ndarray
            (i_1, ..., i_d), if nargout == 2.

        '''
        if np.ndim(x) == 1:
            x = np.reshape(x, [-1, 1])

        y = []
        i = []
        for k in range(self.dim):
            uk = self.X.random_variables[k].cdf(x[:, k])
            y.append(Tensorizer.u2z(uk, self.b, self.d))
            y[k][:, -1] = self.Y.random_variables[k].icdf(y[k][:, -1])
            i.append(y[k][:, :-1])
            y[k] = y[k][:, -1]
        y = np.transpose(y)
        i = np.hstack(i)

        if self.ordering_type == 2:
            j = []
            for k in range(self.d):
                j.append(i[:, np.arange(k, i.shape[1], self.d)])
            i = np.hstack(j)
        elif self.ordering_type != 1:
            raise ValueError('Wrong ordering_type')

        if nargout == 2:
            return y, i
        return np.hstack((i, np.atleast_2d(y)))

    def inverse_map(self, z):
        '''
        Evaluate the map at points z = (i_1, ..., i_d, y), returning x such
        that self.map(x) = z.

        Parameters
        ----------
        z : list or numpy.ndarray
            The points at which the inverse map is to be evaluated.

        Returns
        -------
        numpy.ndarray
            The points x such that self.map(x) = z.

        '''
        z = np.atleast_2d(z)
        u = []
        for k in range(self.dim):
            if self.ordering_type == 1:
                ik = z[:, np.arange(self.d)+k*self.d]
            elif self.ordering_type == 2:
                ik = z[:, np.arange(k, self.dim*self.d, self.dim)]
            else:
                raise ValueError('Wrong ordering_type')
                
            zk = z[:, -self.dim+k]
            zk = self.Y.random_variables[k].cdf(zk)
            u.append(Tensorizer.z2u(np.hstack((ik, np.reshape(zk, [-1, 1]))),
                                    self.b))
            u[k] = self.X.random_variables[k].icdf(u[k])
        return np.transpose(u)

    def tensorize(self, fun):
        '''
        Tensorize a provided function defined on self.X.support().

        Parameters
        ----------
        fun : function or tensap.Function
            The function to be tensorized.

        Raises
        ------
        ValueError
            If the provided argument is neither a tensap.Function nor a
            function.

        Returns
        -------
        tensap.TensorizedFunction
            The tensorized function.

        '''
        if not isinstance(fun, tensap.Function) and \
                not hasattr(fun, '__call__'):
            raise ValueError('The argument must be a tensap.Function or ' +
                             'function.')

        if not isinstance(fun, tensap.Function) and hasattr(fun, '__call__'):
            fun = tensap.UserDefinedFunction(fun, self.dim)
            fun.evaluation_at_multiple_points = True            

        f = tensap.UserDefinedFunction(lambda z: fun(self.inverse_map(z)),
                                       (self.d+1)*self.dim)
        f.evaluation_at_multiple_points = fun.evaluation_at_multiple_points 
        return tensap.TensorizedFunction(f, self)

    def tensorized_function_functional_bases(self, h=1):
        '''
        Return a tensap.FunctionalBases object associated with the provided
        basis or basis function(s) and the Tensorizer object.

        Parameters
        ----------
        h : tensap.FunctionalBases or tensap.FunctionalBasis or function or
        list or scalar, optional
            The function used to generate the basis. The default is the
            function 1.

        Returns
        -------
        tensap.FunctionalBases
            The functional bases.

        '''
        #if isinstance(h, (np.ndarray, list)) or np.isscalar(h):
        #    h = lambda y, h=h: h*np.ones(np.shape(y))
        
        

        if hasattr(h, '__call__'):
            h = tensap.UserDefinedFunctionalBasis([h])
            h.measure = self.Y.random_variables[0]

        if isinstance(h, tensap.FunctionalBasis):
            h = tensap.FunctionalBases.duplicate(h, self.dim)
            
        if np.isscalar(h):
            h = [tensap.PolynomialFunctionalBasis(y.orthonormal_polynomials(),
                                      range(h+1)) for y in self.Y.random_variables]
            h = tensap.FunctionalBases(h)    

        assert isinstance(h, tensap.FunctionalBases), \
            'Wrong type of argument for h.'

        p = tensap.DiscretePolynomials(tensap.DiscreteRandomVariable(
            np.reshape(np.arange(self.b), [-1, 1])))
        p = tensap.PolynomialFunctionalBasis(p, np.arange(self.b))

        bases = [p]*self.d*self.dim + list(h.bases)
        return tensap.FunctionalBases(bases)

    @staticmethod
    def u2z(u, b, d, nargout=1):
        '''
        Return the representation of numbers on [0, 1] in base b with
        resolution d.

        Parameters
        ----------
        u : list or numpy.ndarray
            The inputs to be converted in base b with resolution d.
        b : int
            The base.
        d : int
            The resolution.
        nargout : int, optional
            The number of outputs. The default is 1, returning a horizontal
            stack of i and y. Set to 2 to return y and i separately.

        Returns
        -------
        numpy.ndarray
            Either (i_1, ..., i_d, y) if nargout == 1, or y if nargout == 2.
        numpy.ndarray
            (i_1, ..., i_d), if nargout == 2.

        '''
        u = np.ravel(u)
        su = u*(b**d)
        i = np.floor(su)
        i = np.minimum(i, b**d-1).astype(int)
        y = su - i
        i = tensap.integer2baseb(i, b, d)
        if nargout == 2:
            return y, i
        return np.hstack((i, np.reshape(y, [-1, 1])))

    @staticmethod
    def z2u(z, b):
        '''
        Return the representation of numbers in base b in decimal on [0, 1].

        Parameters
        ----------
        z : list or numpy.ndarray
            The numbers in base b.
        b : int
            The base.

        Returns
        -------
        numpy.ndarray
            The decimal representation of the inputs on [0, 1].

        '''
        z = np.atleast_2d(z)
        d = z.shape[1] - 1
        y = z[:, -1]
        i = tensap.baseb2integer(z[:, :-1].astype(int), b)
        return (y+i)*b**(-d)



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Module is_in.

'''

import numpy as np


def is_in(B, x):
    '''
    Return a boolean array of length x.shape[0] such that ok[k] = True if the
    point x[k, :] is in the box given by B.

    B is of shape (2, dim) and contains the two extreme points B[0, :] and
    B[1, :] of the box.

    Parameters
    ----------
    B : numpy.ndarray
        The box.
    x : numpy.ndarray
        The points.

    Returns
    -------
    ok : numpy.ndarray
        A boolean array indicating if the points x are in the box B.

    '''
    x = np.atleast_2d(x)
    B = np.atleast_2d(B)
    if B.size == 2:
        B = np.reshape(B, [-1, 1])

    ok = np.full(x.shape[0], True)
    for k in range(B.shape[1]):
        ok = np.logical_and(ok, x[:, k] >= B[0, k])
        ok = np.logical_and(ok, x[:, k] <= B[1, k])

    return ok



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Module random_multi_indices.

'''

import numpy as np
import tensap


def random_multi_indices(shape):
    '''
    Return a random variable uniformly distributed on I1 x ... x Id.

    If shape contains integers, the intervals are defined as
    Ij = np.arange(shape[j-1]), j = 1, ..., len(shape).

    Parameters
    ----------
    shape : list or numpy.ndarray
        The number of elements of each interval, or the interval themselves.

    Returns
    -------
    tensap.RandomVector
        The random variable uniformly distributed on I1 x ... x Id.

    '''
    order = len(shape)

    if np.all([isinstance(x, (list, np.ndarray)) for x in shape]):
        ind = shape
    else:
        ind = []
        for dim in range(order):
            ind.append(np.arange(shape[dim]))

    for dim in range(order):
        ind[dim] = tensap.DiscreteRandomVariable(ind[dim])

    return tensap.RandomVector(ind)



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Module discrete_random_variable.

'''

import numpy as np
import tensap


class DiscreteRandomVariable(tensap.RandomVariable):
    '''
    Class DiscreteRandomVariable.

    Attributes
    ----------
    values : list or numpy.ndarray
        Array containing the set of values (x_1,...,x_N) taken by the random
        variable, with x_i in R^d, i = 1, ..., N.
    probabilities : list or numpy.ndarray, optional
        Array containing the probabilities P(X=x_1), ..., P(X=x_N). The default
        is None, indicating to create a uniform law: P(X=x_i)=1/N,
        i = 1, ..., N.

    '''

    def __init__(self, values, probabilities=None):
        '''
        Discrete random variable taking a finite set of values in R^d.

        Parameters
        ----------
        values : list or numpy.ndarray
            Array containing the set of values (x_1,...,x_N) taken by the
            random variable, with x_i in R^d, i = 1, ..., N.
        probabilities : list or numpy.ndarray, optional
            Array containing the probabilities P(X=x_1), ..., P(X=x_N). The
            default is None, indicating to create a uniform law: P(X=x_i)=1/N,
            i = 1, ..., N.

        Raises
        ------
        ValueError
            If the provided probabilities are not authorized.

        Returns
        -------
        None.

        '''
        tensap.RandomVariable.__init__(self)

        if np.ndim(values) == 1:
            values = np.reshape(values, (-1, 1))

        self.values = values
        N = values.shape[0]
        if probabilities is None:
            probabilities = np.full(N, 1/N)
        elif len(probabilities) != N:
            raise ValueError('The arguments must have the same size.')
        elif not np.all([x >= 0 for x in probabilities]):
            raise ValueError('All the probabilities should be >= 0.')

        self.probabilities = np.array(probabilities)
        if np.abs(np.sum(self.probabilities) - 1) > np.finfo(float).eps:
            self.probabilities = self.probabilities / \
                np.sum(self.probabilities)

    def get_standard_random_variable(self):
        '''
        Return the standard discrete random variable.

        Returns
        -------
        tensap.UniformRandomVariable
            The standard discrete random variable.

        '''
        values = np.array(self.values)
        values = (values - np.tile(self.mean(), (values.shape[0], 1))) / \
            np.tile(self.std(), (values.shape[0], 1))
        return DiscreteRandomVariable(values, self.probabilities)

    def support(self):
        return np.vstack((np.min(self.values, 0), np.max(self.values, 0)))

    def plot(self, quantity, *args):
        '''
        Plot the desired quantity, chosen between 'pdf', 'cdf' or 'icdf'.

        Parameters
        ----------
        quantity : str
            The desired quantity, chosen between 'pdf', 'cdf' or 'icdf'.
        *args : tuple
            Additional parameters for matplotlib.pyplot's function step (for
            the cdf), vlines (for the pdf) or plot (for the icdf).

        Raises
        ------
        ValueError
            If the provided argument quantity is wrong.

        Returns
        -------
        None.

        '''
        import matplotlib.pyplot as plt

        x = np.ravel(self.values)
        y = self.probabilities
        delta = np.max(x) - np.min(x)
        ax = [np.min(x)-delta/10, np.max(x)+delta/10]

        if quantity == 'cdf':
            x = np.concatenate(([ax[0]], x, [ax[1]]))
            y = np.concatenate(([0, 0], np.cumsum(y)))
            plt.step(x, y, *args)
            plt.xlim(*ax)
            plt.ylim(0, 1.1*np.max(y))
        elif quantity == 'pdf':
            plt.vlines(x, np.zeros(y.shape), y, *args)
            plt.xlim(*ax)
            plt.ylim(0, 1.1*np.max(y))
        elif quantity == 'icdf':
            u = np.linspace(0, 1, 500)
            y = self.icdf(u)
            plt.plot(u, y, *args)
            plt.xlim(0, 1)
            plt.ylim(*ax)
        else:
            raise ValueError('Wront argument value.')
        plt.show()

    def integration_rule(self):
        '''
        Return the integration rule object associated with the discrete random
        variable.

        Returns
        -------
        tensap.IntegrationRule
            The integration rule object associated with the discrete random
            variable.

        '''
        return tensap.IntegrationRule(self.values, self.probabilities)

    def get_parameters(self):
        return self.values, self.probabilities

    def cdf(self, x):
        '''
        Compute the cumulative density function of X at x.

        Parameters
        ----------
        x : float, list or numpy.ndarray
            The points at which the cumulative density function of X is to be
            evaluated.

        Returns
        -------
        numpy.ndarray
            The evaluations of the cumulative density function of X at x.

        '''
        x = np.array(x)
        shape = x.shape
        M = x.size

        x = np.tile(np.reshape(x, (1, M)), (self.values.size, 1))
        v = np.tile(np.reshape(self.values, (-1, 1)), (1, M))
        w = np.tile(np.reshape(self.probabilities, (-1, 1)), (1, M))
        w[v > x] = 0
        return np.reshape(np.sum(w, 0), shape)

    def icdf(self, p):
        '''
        Compute the inverse cumulative density function of X at p (quantile).

        Parameters
        ----------
        p : float, list or numpy.ndarray
            The points at which the inverse cumulative density function of X is
            to be evaluated.

        Returns
        -------
        numpy.ndarray
            The evaluations of the inverse cumulative density function of X at
            x.

        '''
        p = np.array(p)
        shape = p.shape
        v = np.vstack(([-np.inf], np.reshape(self.values, (-1, 1))))
        F = np.vstack(([0], np.reshape(self.probabilities, (-1, 1))))

        M = p.size
        p = np.tile(np.reshape(p, (1, M)), (F.size, 1))
        F = np.tile(np.reshape(np.cumsum(F), (-1, 1)), (1, M))
        return np.reshape(v[np.nonzero(np.cumsum(np.transpose(F >= p), 1) ==
                                       1)[1]], shape)

    def mean(self):
        return np.matmul(np.transpose(self.probabilities), self.values)

    def var(self):
        '''
        Return the variance of the random variable.

        Returns
        -------
        float
            The variance of the random variable.

        '''
        return np.matmul(np.transpose(self.probabilities), self.values**2) - \
            self.mean()**2

    def random_variable_statistics(self):
        '''
        Return the mean and the variance of the discrete random variable.

        Returns
        -------
        float
            The mean of the random variable.
        float
            The variance of the random variable.

        '''
        return self.mean(), self.var()

    def random(self, n=1):
        '''
        Generate n random numbers according to the distribution of the
        RandomVariable.

        Parameters
        ----------
        n : int
            The number of random numbers generated.

        Returns
        -------
        numpy.ndarray
            The generated numbers.

        '''
        Y = DiscreteRandomVariable(np.arange(len(self.probabilities)),
                                   self.probabilities)
        ind = Y.icdf(np.random.rand(int(n)))
        return np.squeeze(self.values[ind.astype(int), :], 1)
